@@ -44,6 +44,10 @@
 //! assert_ne!(x, y);
 //! ```
 
+// Enable the `doc_cfg` feature when the `docsrs` configuration attribute is
+// defined
+#![cfg_attr(docsrs, feature(doc_cfg))]
+
 #[cfg(test)]
 #[cfg(feature = "tinyset")]
 use quickcheck::quickcheck;
@@ -54,6 +58,7 @@ use boxedset::HashSet;
 use dashmap::{mapref::entry::Entry, DashMap};
 use parking_lot::Mutex;
 use std::cell::RefCell;
+#[cfg(feature = "arc")]
 use std::sync::atomic::AtomicUsize;
 #[cfg(feature = "arc")]
 use std::sync::atomic::Ordering;
@@ -61,9 +66,8 @@ use std::sync::atomic::Ordering;
 mod container;
 use container::{TypeHolder, TypeHolderSend};
 
-use std::any::Any;
-#[cfg(feature = "arc")]
 use std::any::TypeId;
+use std::any::Any;
 use std::borrow::Borrow;
 use std::convert::AsRef;
 use std::fmt::{Debug, Display, Pointer};
@@ -134,6 +138,8 @@ fn like_doctest_localintern() {
     assert_eq!(&*x, "hello"); // dereference a Intern like a pointer\
 }
 
+/// A pointer to an interned object that has been leaked and may be used in any
+/// thread without locking.
 pub struct Intern<T: 'static> {
     pointer: &'static T,
 }
@@ -195,9 +201,82 @@ where
     F: FnOnce(&mut T) -> R,
     T: Any + Send + Default,
 {
-    static INTERN_CONTAINERS: Mutex<TypeHolderSend> =
-        parking_lot::const_mutex(TypeHolderSend::new());
-    f(INTERN_CONTAINERS.lock().get_type_mut())
+    // Compute the hash of the type.
+    fn hash_of_type<T: 'static>() -> u64 {
+        // We use very simple hasher, because it is optimized away to a constant:
+        // https://rust.godbolt.org/z/4T1fa4GGs
+        // which is not true for using `DefaultHasher`:
+        // https://rust.godbolt.org/z/qKar1WKfz
+        struct HasherForTypeId {
+            hash: u64,
+        }
+
+        impl Hasher for HasherForTypeId {
+            fn write(&mut self, bytes: &[u8]) {
+                // Hash for type only calls `write_u64` once,
+                // but handle this case explicitly to make sure
+                // this code doesn't break if stdlib internals change.
+
+                for byte in bytes {
+                    self.hash = self.hash.wrapping_mul(31).wrapping_add(*byte as u64);
+                }
+            }
+
+            fn write_u64(&mut self, v: u64) {
+                self.hash = v;
+            }
+
+            fn finish(&self) -> u64 {
+                self.hash
+            }
+        }
+
+        let mut hasher = HasherForTypeId { hash: 0 };
+        TypeId::of::<T>().hash(&mut hasher);
+        hasher.hash
+    }
+
+    const INTERN_CONTAINER_COUNT: usize = 32;
+    static INTERN_CONTAINERS: [Mutex<TypeHolderSend>; INTERN_CONTAINER_COUNT] = [
+        // There's no way to create a static array without copy-paste.
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+        parking_lot::const_mutex(TypeHolderSend::new()),
+    ];
+    f(
+        INTERN_CONTAINERS[hash_of_type::<T>() as usize % INTERN_CONTAINER_COUNT]
+            .lock()
+            .get_type_mut(),
+    )
 }
 
 impl<T: Eq + Hash + Send + Sync + 'static> Intern<T> {
@@ -246,6 +325,12 @@ impl<T: Eq + Hash + Send + Sync + 'static> Intern<T> {
     pub fn num_objects_interned() -> usize {
         with_global(|m: &mut HashSet<&'static T>| -> usize { m.len() })
     }
+
+    /// Only for benchmarking, this will cause problems
+    #[cfg(feature = "bench")]
+    pub fn benchmarking_only_clear_interns() {
+        with_global(|m: &mut HashSet<&'static T>| -> () { m.clear() })
+    }
 }
 
 #[cfg(feature = "tinyset")]
@@ -268,9 +353,7 @@ fn heap_location() -> u64 {
             std::sync::atomic::Ordering::Relaxed,
             std::sync::atomic::Ordering::Relaxed,
         ) {
-            Ok(_) => {
-                ptr as u64
-            },
+            Ok(_) => ptr as u64,
             Err(ptr) => {
                 println!("race, ptr is {:p}", ptr);
                 ptr as u64
@@ -290,6 +373,7 @@ const fn sz<T>() -> u64 {
 /// of the resulting u64 be zero, which will mean that `Set64` (which is
 /// space-efficient in storing small integers) can store this result in far
 /// fewer than 8 bytes.
+#[cfg_attr(docsrs, doc(cfg(feature = "tinyset")))]
 #[cfg(feature = "tinyset")]
 impl<T: Debug> Fits64 for Intern<T> {
     unsafe fn from_u64(x: u64) -> Self {
@@ -307,6 +391,7 @@ impl<T: Debug> Fits64 for Intern<T> {
 /// significant bits of the resulting u64 be zero, which will mean
 /// that `Set64` (which is space-efficient in storing small integers)
 /// can store this result in fewer than 8 bytes.
+#[cfg_attr(docsrs, doc(cfg(feature = "tinyset")))]
 #[cfg(feature = "tinyset")]
 impl<T: Debug> Fits64 for LocalIntern<T> {
     unsafe fn from_u64(x: u64) -> Self {
@@ -377,6 +462,7 @@ fn test_intern_set64() {
 /// assert_eq!(x, ArcIntern::from("hello"));
 /// assert_eq!(&*x, "hello"); // dereference an ArcIntern like a pointer
 /// ```
+#[cfg_attr(docsrs, doc(cfg(feature = "arc")))]
 #[cfg(feature = "arc")]
 pub struct ArcIntern<T: Eq + Hash + Send + Sync + 'static> {
     pointer: std::ptr::NonNull<RefCount<T>>,
@@ -387,23 +473,28 @@ unsafe impl<T: Eq + Hash + Send + Sync> Send for ArcIntern<T> {}
 #[cfg(feature = "arc")]
 unsafe impl<T: Eq + Hash + Send + Sync> Sync for ArcIntern<T> {}
 
+#[cfg(feature = "arc")]
 #[derive(Debug)]
 struct RefCount<T> {
     count: AtomicUsize,
     data: T,
 }
+#[cfg(feature = "arc")]
 impl<T: Eq> Eq for RefCount<T> {}
+#[cfg(feature = "arc")]
 impl<T: PartialEq> PartialEq for RefCount<T> {
     fn eq(&self, other: &Self) -> bool {
         self.data == other.data
     }
 }
+#[cfg(feature = "arc")]
 impl<T: Hash> Hash for RefCount<T> {
     fn hash<H: Hasher>(&self, hasher: &mut H) {
         self.data.hash(hasher)
     }
 }
 
+#[cfg(feature = "arc")]
 #[derive(Eq, PartialEq, Hash)]
 struct BoxRefCount<T>(Box<RefCount<T>>);
 #[cfg(feature = "arc")]
@@ -412,16 +503,19 @@ impl<T> BoxRefCount<T> {
         self.0.data
     }
 }
+#[cfg(feature = "arc")]
 impl<T> Borrow<T> for BoxRefCount<T> {
     fn borrow(&self) -> &T {
         &self.0.data
     }
 }
+#[cfg(feature = "arc")]
 impl<T> Borrow<RefCount<T>> for BoxRefCount<T> {
     fn borrow(&self) -> &RefCount<T> {
         &self.0
     }
 }
+#[cfg(feature = "arc")]
 impl<T> Deref for BoxRefCount<T> {
     type Target = T;
     fn deref(&self) -> &Self::Target {
@@ -534,6 +628,10 @@ impl<T: Eq + Hash + Send + Sync + 'static> ArcIntern<T> {
     pub fn refcount(&self) -> usize {
         unsafe { self.pointer.as_ref().count.load(Ordering::Acquire) }
     }
+
+    /// Only for benchmarking, this will cause problems
+    #[cfg(feature = "bench")]
+    pub fn benchmarking_only_clear_interns() {}
 }
 
 #[cfg(feature = "arc")]
@@ -684,6 +782,7 @@ macro_rules! create_impls {
             }
         }
 
+        #[cfg_attr(docsrs, doc(cfg(feature = "serde")))]
 		#[cfg(feature = "serde")]
 		impl<T: $( $traits +)* Serialize> Serialize for $Intern<T> {
             fn serialize<S: Serializer>(&self, serializer: S) -> Result<S::Ok, S::Error> {
@@ -691,6 +790,7 @@ macro_rules! create_impls {
             }
         }
 
+        #[cfg_attr(docsrs, doc(cfg(feature = "serde")))]
 		#[cfg(feature = "serde")]
 		impl<'de, T: $( $newtraits +)* 'static + Deserialize<'de>> Deserialize<'de> for $Intern<T> {
             fn deserialize<D: Deserializer<'de>>(deserializer: D) -> Result<Self, D::Error> {
@@ -1003,6 +1103,10 @@ impl<T: Eq + Hash + 'static> LocalIntern<T> {
     pub fn num_objects_interned() -> usize {
         with_local(|m: &mut HashSet<Box<T>>| -> usize { m.len() })
     }
+
+    /// Only for benchmarking, this will cause problems
+    #[cfg(feature = "bench")]
+    pub fn benchmarking_only_clear_interns() {}
 }
 
 #[test]
